"""
Module used for defining the certificate class and it's functionality.
"""

import sys
import random
import string
from datetime import datetime, timedelta
from typing import Union
from os.path import abspath, dirname, join
from cryptography.hazmat.primitives.asymmetric import rsa, ec
from cryptography.hazmat.primitives import serialization
from .SimUtils import hash_info

script_dir = dirname(abspath(__file__))

if script_dir in ['PKI_Practice', 'PKI Practice', 'app']:
    sys.path.append(abspath(script_dir))
elif script_dir == 'PKIPractice':
    sys.path.append(abspath(join(script_dir, '..')))
else:
    sys.path.append(abspath(join(script_dir, '../..')))

from PKIPractice.Utilities.DataclassUtils import *


class PKICertificate:
    """
    The PKICertificate class represents a given certificate that can be sent and held by holders.

    Attributes:
        name (str): The name of the certificate, derived from the `subject_name` parameter.
        subject_info (SUBJECT_INFO): Information about the subject, including common name, country,
            state, organization, and contact details.
        issuer_info (ISSUER_INFO): Information about the certificate issuer, such as common name,
            country, state, and organization details.
        serial_number (str): A unique 16-character hexadecimal string generated for the certificate.
        x509_version (int): The X.509 version of the certificate, defaulting to version 3.
        signature_alg (str): The signature algorithm used for signing the certificate, derived from
            `subject_env_info.sig_hash`.
        valid_start (datetime): The datetime when the certificate becomes valid, initialized to the
            current time.
        valid_end (Union[datetime, None]): The datetime when the certificate expires. If
            `subject_env_info.cert_valid_dur` is `'none'`, this is `None`. Otherwise, it is computed
            based on the duration string (formatted as `HH:MM:SS`).
        encryption_alg (dict): Details of the encryption algorithm used in the certificate, derived
            from `subject_env_info.encrypt_alg`.
        subject_pub_key (Union[rsa.RSAPublicKey, ec.EllipticCurvePublicKey]): The subject's public key,
            which can be either an RSA or elliptic curve public key.
        signature (str): The certificate's digital signature, generated by hashing the certificate
            content using the specified `signature_alg`.

    Methods:
        hash_content() -> str:
            Returns the core attributes of the certificate as a string.
    """
    def __init__(self, subject_name: str, subject_info: HOLDER_INFO,
                 issuer_info: HOLDER_INFO, subject_env_info: HOLDER_ENV_INFO,
                 subject_pub_key: Union[rsa.RSAPublicKey, ec.EllipticCurvePublicKey]):

        # Name
        self.name = subject_name

        # Subject info
        self.subject_info: SUBJECT_INFO = SUBJECT_INFO(
            common_name=subject_info.common_name, country=subject_info.country, state=subject_info.state,
            local=subject_info.local, org=subject_info.org, org_unit=subject_info.org_unit, email=subject_info.email,
            url=subject_info.url
        )

        # Issuer info
        self.issuer_info: ISSUER_INFO = ISSUER_INFO(
            common_name=issuer_info.common_name, country=issuer_info.country, state=issuer_info.state,
            local=issuer_info.local, org=issuer_info.org, url=issuer_info.url
        )

        # Certificate information
        self.serial_number: str = ''.join(random.choices(string.hexdigits[:16], k=16)).upper()
        self.x509_version: int = 3
        self.signature_alg: str = subject_env_info.sig_hash

        # Time information
        self.valid_start: datetime = datetime.now()
        if subject_env_info.cert_valid_dur == 'none':
            self.valid_end: Union[datetime, None] = None
        else:
            hours, minutes, seconds = map(int, subject_env_info.cert_valid_dur.split(":"))
            delta = timedelta(hours=hours, minutes=minutes, seconds=seconds)
            self.valid_end: Union[datetime, None] = self.valid_start + delta

        # Asymmetric Key information
        self.encryption_alg: dict = subject_env_info.encrypt_alg
        self.subject_pub_key: Union[rsa.RSAPublicKey, ec.EllipticCurvePublicKey] = subject_pub_key

        self.signature: str = hash_info(self.hash_content, self.signature_alg)

    @property
    def hash_content(self) -> str:
        """
        Returns the core attributes of the certificate as a hashable string.

        Returns:
            str: The core attributes turned into string and lined up.
        """

        subject_pub_pem = self.subject_pub_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )

        return f'{self.name}{self.subject_info}{self.issuer_info}' \
               f'{self.serial_number}{self.x509_version}{self.signature_alg}' \
               f'{self.valid_start}{self.valid_end}{self.encryption_alg}{subject_pub_pem}'



"""
This module provides utilities for parsing configuration files and working with enums.

Utilities:
    parse_config_auto: Parses an autoconfiguration file in JSON, YAML, XML, or TOML format.
    parse_config_manual: Parses a manual configuration file in JSON, YAML, XML, or TOML format.
    EnumUtils: Utilities for working with enums.
"""

__all__ = ["IngestUtils", "EnumUtils", "CLIUtils", "DataclassUtils"]

from . import EnumUtils, CLIUtils, IngestUtils, DataclassUtils



"""
Module for testing configuration ingestion utilities.
"""

import unittest

# Relative pathing from project root
import sys
from os.path import curdir, abspath, basename, dirname, join

script_dir = dirname(abspath(__file__))

if script_dir in ['PKI_Practice', 'PKI Practice', 'app']:
    sys.path.append(abspath(script_dir))
elif script_dir == 'PKIPractice':
    sys.path.append(abspath(join(script_dir, '..')))
else:
    sys.path.append(abspath(join(script_dir, '../..')))

# Personal Modules must be imported after the system path is modified.
from PKIPractice.Utilities.IngestUtils import parse_config_auto, parse_config_manual


class TestIngestion(unittest.TestCase):
    def setUp(self) -> None:
        """
        Sets up the parameters for testing.
        """

        current_dir = basename(abspath(curdir))
        if current_dir in ['PKI_Practice', 'PKI Practice', 'app']:
            self.dc_dir = './'
        elif current_dir == 'PKIPractice':
            self.dc_dir = '../'
        elif current_dir == 'tests':
            self.dc_dir = '../../'
        else:
            self.dc_dir = './'
            
    def test_key_count_auto(self) -> None:
        """
        Checks if the number of keys in the auto config files is consistent across all formats.
        """

        def total_keys(test_dict: dict) -> int:
            """
            Returns how many keys are in the dictionary.
            """

            return 0 if not isinstance(test_dict, dict) else len(test_dict) + sum(
                total_keys(val) for val in test_dict.values())

        config_json: dict = parse_config_auto(f'{self.dc_dir}Default_Configs/default_auto.json')
        config_xml: dict = parse_config_auto(f'{self.dc_dir}Default_Configs/default_auto.xml')
        config_toml: dict = parse_config_auto(f'{self.dc_dir}Default_Configs/default_auto.toml')

        count_json: int = total_keys(config_json)
        count_xml: int = total_keys(config_xml)
        count_toml: int = total_keys(config_toml)

        self.assertNotEqual(count_json, 0, 'No keys found when parsing auto JSON.')
        self.assertNotEqual(count_xml, 0, 'No keys found when parsing auto XML.')
        self.assertNotEqual(count_toml, 0, 'No keys found when parsing auto TOML.')

        self.assertEqual(count_json, count_xml, 'auto JSON and auto XML key count are not equal.')
        self.assertEqual(count_json, count_toml, 'auto JSON and auto TOML key count are not equal.')

        # Accommodating for earlier versions before Python 3.10
        if sys.version_info[1] > 9:
            config_yaml: dict = parse_config_auto(f'{self.dc_dir}Default_Configs/default_auto.yaml')
            count_yaml: int = total_keys(config_yaml)
            self.assertNotEqual(count_yaml, 0, 'No keys found when parsing auto YAML.')
            self.assertEqual(count_json, count_yaml, 'auto JSON and auto YAML key count are not equal.')

    def test_key_count_manual(self) -> None:
        """
        Checks if the number of keys in the manual config files is consistent across all formats.
        """
        def total_keys(test_dict: dict) -> int:
            """
            Returns how many keys are in the dictionary.
            """
            return 0 if not isinstance(test_dict, dict) else len(test_dict) + sum(
                total_keys(val) for val in test_dict.values())

        config_json: dict = parse_config_manual(f'{self.dc_dir}Default_Configs/default_manual.json')
        config_xml: dict = parse_config_manual(f'{self.dc_dir}Default_Configs/default_manual.xml')
        config_toml: dict = parse_config_manual(f'{self.dc_dir}Default_Configs/default_manual.toml')

        count_json: int = total_keys(config_json)
        count_xml: int = total_keys(config_xml)
        count_toml: int = total_keys(config_toml)

        self.assertNotEqual(count_json, 0, 'No keys found when parsing manual JSON.')
        self.assertNotEqual(count_xml, 0, 'No keys found when parsing manual XML.')
        self.assertNotEqual(count_toml, 0, 'No keys found when parsing manual TOML.')

        self.assertEqual(count_json, count_xml, 'manual JSON and manual XML key count are not equal.')
        self.assertEqual(count_json, count_toml, 'manual JSON and manual TOML key count are not equal.')

        # Accommodating for earlier versions before Python 3.10
        if sys.version_info[1] > 9:
            config_yaml: dict = parse_config_manual(f'{self.dc_dir}Default_Configs/default_manual.yaml')
            count_yaml: int = total_keys(config_yaml)
            self.assertNotEqual(count_yaml, 0, 'No keys found when parsing manual YAML.')
            self.assertEqual(count_json, count_yaml, 'manual JSON and manual YAML key count are not equal.')

    def test_key_types_auto(self) -> None:
        """
        Checks if the key types in the auto config files are consistent across all formats.
        """
        config_json: dict = parse_config_auto(f'{self.dc_dir}Default_Configs/default_auto.json')
        config_xml: dict = parse_config_auto(f'{self.dc_dir}Default_Configs/default_auto.xml')
        config_toml: dict = parse_config_auto(f'{self.dc_dir}Default_Configs/default_auto.toml')

        del config_json['log_save_filepath']
        del config_xml['log_save_filepath']
        del config_toml['log_save_filepath']

        self.assertEqual(config_json, config_xml, 'auto JSON and auto XML are not equal.')
        self.assertEqual(config_json, config_toml, 'auto JSON and auto TOML are not equal.')

        # Accommodating for earlier versions before Python 3.10
        if sys.version_info[1] > 9:
            config_yaml: dict = parse_config_auto(f'{self.dc_dir}Default_Configs/default_auto.yaml')

            del config_yaml['log_save_filepath']
            self.assertEqual(config_json, config_yaml, 'auto JSON and auto YAML are not equal.')

    def test_key_types_manual(self) -> None:
        """
        Checks if the key types in the manual config files are consistent across all formats.
        """
        config_json: dict = parse_config_manual(f'{self.dc_dir}Default_Configs/default_manual.json')
        config_xml: dict = parse_config_manual(f'{self.dc_dir}Default_Configs/default_manual.xml')
        config_toml: dict = parse_config_manual(f'{self.dc_dir}Default_Configs/default_manual.toml')

        self.assertEqual(config_json, config_xml, 'manual JSON and manual XML are not equal.')
        self.assertEqual(config_json, config_toml, 'manual JSON and manual TOML are not equal.')

        # Accommodating for earlier versions before Python 3.10
        if sys.version_info[1] > 9:
            config_yaml: dict = parse_config_manual(f'{self.dc_dir}Default_Configs/default_manual.yaml')
            self.assertEqual(config_json, config_yaml, 'manual JSON and manual YAML are not equal.')


if __name__ == '__main__':
    unittest.main()



"""
Package containing all tests for the package.

Modules:
    test_enums - tests all enumerations and functionality for enumerations.
    test_ingestion - tests the functionality of argument passing using default files.
    test_cli - tests the command line interface and the options that can be passed.
"""

__all__ = ["test_enums", "test_ingestion", "test_cli"]

from . import test_enums
from . import test_ingestion
from . import test_cli



"""
Module for testing the network class.
"""

import unittest
import tempfile
from datetime import datetime
from typing import List

# Relative pathing from project root
import sys
from os.path import abspath, dirname, join, exists

script_dir = dirname(abspath(__file__))

if script_dir in ['PKI_Practice', 'PKI Practice', 'app']:
    sys.path.append(abspath(script_dir))
elif script_dir == 'PKIPractice':
    sys.path.append(abspath(join(script_dir, '..')))
else:
    sys.path.append(abspath(join(script_dir, '../..')))

# Personal Modules must be imported after the system path is modified.
from PKIPractice.Utilities.CLIUtils import ingest_config
from PKIPractice.Simulation.Network import PKINetwork
from PKIPractice.Simulation.Holder import PKIHolder


class TestNetwork(unittest.TestCase):
    def setUp(self) -> None:
        """
        Sets up the parameters for testing.
        """

        self.env_auto_settings, self.env_manual_settings = ingest_config(sys.argv, default=True)

    def test_network_creation(self) -> None:
        """
        Tests the ability to create a whole network object.
        """

        pki_network: PKINetwork = PKINetwork('Test_Net', self.env_auto_settings)
        self.assertIsNotNone(pki_network, 'Something went wrong when generating autoconfiguration.')

        pki_network: PKINetwork = PKINetwork('Test_Net', self.env_auto_settings, self.env_manual_settings)
        self.assertIsNotNone(pki_network, 'Something went wrong when generating auto and manual configuration.')

    def test_holders(self) -> None:
        """
        Tests the creation and organization of holders.
        """

        auto_network: PKINetwork = PKINetwork('Test_Net', self.env_auto_settings)
        manual_network: PKINetwork = PKINetwork('Test_Net', self.env_auto_settings, self.env_manual_settings)

        for pki_network in [auto_network, manual_network]:
            all_holders: List[PKIHolder] = pki_network.get_network()

            # Assert that all holders were created
            self.assertEqual(
                sum(pki_network.network_count_by_level),
                len(all_holders),
                'All required holders were not created (by level count).'
            )
            self.assertEqual(
                len(all_holders),
                pki_network.network_total_count,
                'All required holders were not created (by total count).'
            )

            # Assert the right amount of holders were created per level
            for i in range(pki_network.network_level_count):
                self.assertEqual(
                    pki_network.network_count_by_level[i],
                    len(pki_network.network[i+1]),
                    f'All required holders for level {i+1} were not created.'
                )

            # Assert the top level of holders are all root CAs
            for holder in pki_network.network[1]:
                self.assertTrue(
                    holder.holder_type_info.ca_status == 'root_auth',
                    'All first level holders are not root CAs.'
                )

            # Assert all levels but the first and last are intermediate CAs
            for i in range(1, pki_network.network_level_count - 1):
                for holder in pki_network.network[i+1]:
                    self.assertTrue(
                        holder.holder_type_info.ca_status == 'inter_auth',
                        f'All level {i+1} holders are not intermediary CAs.'
                    )

            # Assert the bottom level of holders are not CAs
            for holder in pki_network.network[pki_network.network_level_count]:
                self.assertTrue(
                    holder.holder_type_info.ca_status == 'not_auth',
                    'All last level holders are not regular holders.'
                )

    def test_log_output(self) -> None:
        """
        Tests the ability to create a proper log file.
        """

        with tempfile.TemporaryDirectory() as temp_dir:
            # Set filepath
            new_fp: str = join(temp_dir, self.env_auto_settings["log_save_filepath"])
            self.env_auto_settings["log_save_filepath"] = new_fp

            # Create the network
            pki_network: PKINetwork = PKINetwork('Test Net With Space', self.env_auto_settings)
            pki_network.save_logs()

            # Run tests on temporary file
            self.assertTrue(exists(new_fp), 'The test file was not created properly.')

            with open(new_fp, 'r') as f:
                output: List[str] = f.readlines()

                # Compact tests
                # Header row
                self.assertEqual(
                    output[0],
                    'ID, Timestamp, Category, Success, Subject, Act, Output, Origin, Message\n',
                    'The expected header row was not found.'
                )

                # New lines
                self.assertTrue(all(line[-1:] == '\n' for line in output), 'All lines do not end with a new line.')

                # Right amount of commas
                self.assertTrue(
                    all(line.count(',') == 8 for line in output),
                    'All lines do not have the right amonut of commas.'
                )

                # Detailed tests
                for line in output:
                    row: List[str] = line.split(',')

                    if row[0] == 'ID':
                        continue

                    test_successful: bool = True
                    try:
                        # Convertable ID
                        self.assertIsInstance(
                            int(row[0]), int,
                            f'The first value in row {row} could not be read as an integer.'
                        )

                        # Proper timestamp
                        self.assertIsInstance(
                            datetime.strptime(row[1], ' %Y-%m-%d %H:%M:%S.%f'),
                            datetime,
                            f'The timestamp in row {row} could not be read as a proper datetime object.'
                        )

                        # No spaces in one-word items
                        self.assertTrue(
                            all(" " not in item.strip() for item in row[2:7]),
                            f'There were spaces found in one-word columns for row {row}'
                        )
                    except (TypeError, ValueError):
                        test_successful = False

                    # Everything passed without an exception
                    self.assertTrue(test_successful, f'Item-wise row tests did not pass safely for row {row}')


if __name__ == "__main__":
    unittest.main()



"""
Testing module for command line interface.
"""

import unittest
import subprocess
import random
import string
from typing import List

# Relative pathing from project root
import sys
from os.path import abspath, dirname, join, basename, curdir

script_dir = dirname(abspath(__file__))

if script_dir in ['PKI_Practice', 'PKI Practice', 'app']:
    sys.path.append(abspath(script_dir))
elif script_dir == 'PKIPractice':
    sys.path.append(abspath(join(script_dir, '..')))
else:
    sys.path.append(abspath(join(script_dir, '../..')))


class TestCLI(unittest.TestCase):
    def setUp(self) -> None:
        """
        Set up the test environment.
        """
        current_dir = basename(abspath(curdir))
        if current_dir in ['PKI_Practice', 'PKI Practice', 'app']:
            self.dc_dir = './'
            self.pyfile = './PKIPractice/RunConfig.py'
        elif current_dir == 'PKIPractice':
            self.dc_dir = '../'
            self.pyfile = './RunConfig.py'
        elif current_dir == 'tests':
            self.dc_dir = '../../'
            self.pyfile = '../RunConfig.py'
        else:
            self.dc_dir = './'
            self.pyfile = './PKIPractice/RunConfig.py'

    def test_help(self) -> None:
        """
        Test the help flag.
        """
        result = subprocess.run(['python', self.pyfile, '-t', '-h'], capture_output=True)
        print(result)

        self.assertEqual(0, result.returncode, 'Test Help returned an invalid return code.')
        self.assertEqual(b'', result.stderr, 'Test Help returned an error message.')

        result = subprocess.run(['python', self.pyfile, '-t', '--help'], capture_output=True)
        print(result)

        self.assertEqual(0, result.returncode, 'Test Help returned an invalid return code.')
        self.assertEqual(b'', result.stderr, 'Test Help returned an error message.')

        result = subprocess.run(['python', self.pyfile, '-h'], capture_output=True)
        print(result)

        self.assertEqual(0, result.returncode, 'Help returned an invalid return code.')
        self.assertEqual(b'', result.stderr, 'Help returned an error message.')

        result = subprocess.run(['python', self.pyfile, '-h', '--help'], capture_output=True)
        print(result)

        self.assertEqual(0, result.returncode, 'Double Help returned an invalid return code.')
        self.assertEqual(b'', result.stderr, 'Double Help returned an error message.')

    def test_default(self) -> None:
        """
        Test the default flag.
        """
        result = subprocess.run(['python', self.pyfile, '-t', '-d'], capture_output=True)
        print(result)

        self.assertEqual(0, result.returncode, 'Test Default returned an invalid return code.')
        self.assertEqual(b'', result.stderr, 'Test Default returned an error message.')

        result = subprocess.run(['python', self.pyfile, '-t', '--default'], capture_output=True)
        print(result)

        self.assertEqual(0, result.returncode, 'Test Default returned an invalid return code.')
        self.assertEqual(b'', result.stderr, 'Test Default returned an error message.')

        result = subprocess.run(['python', self.pyfile, '-t', '-d', '--default'], capture_output=True)
        print(result)

        self.assertEqual(0, result.returncode, 'Test Double Default returned an invalid return code.')
        self.assertEqual(b'', result.stderr, 'Test Double Default returned an error message.')

    def test_args(self) -> None:
        """
        Test the arguments using files from Default_Configs folder.
        """
        arg_combos = [
            ['python', self.pyfile, '-t', f'{self.dc_dir}Default_Configs/default_auto.yaml'],
            ['python', self.pyfile, '-t', f'{self.dc_dir}Default_Configs/default_auto.json'],
            ['python', self.pyfile, '-t', f'{self.dc_dir}Default_Configs/default_auto.toml'],
            ['python', self.pyfile, '-t', f'{self.dc_dir}Default_Configs/default_auto.xml'],
            [
                'python',
                self.pyfile, '-t',
                f'{self.dc_dir}Default_Configs/default_auto.yaml',
                f'{self.dc_dir}Default_Configs/default_manual.yaml'
            ],
            [
                'python',
                self.pyfile, '-t',
                f'{self.dc_dir}Default_Configs/default_auto.json',
                f'{self.dc_dir}Default_Configs/default_manual.json'
            ],
            [
                'python',
                self.pyfile, '-t',
                f'{self.dc_dir}Default_Configs/default_auto.toml',
                f'{self.dc_dir}Default_Configs/default_manual.toml'
            ],
            [
                'python',
                self.pyfile, '-t',
                f'{self.dc_dir}Default_Configs/default_auto.xml',
                f'{self.dc_dir}Default_Configs/default_manual.xml'
            ]
        ]

        for args in arg_combos:
            result = subprocess.run(args, capture_output=True)
            print(result)

            self.assertEqual(
                0,
                result.returncode,
                f'Failed with args: {args}. {dirname(abspath(__file__))} Full file path: {abspath(self.pyfile)}'
            )
            self.assertEqual(
                b'',
                result.stderr,
                f'Failed with args: {args}. {dirname(abspath(__file__))} Full file path: {abspath(self.pyfile)}'
            )

    def test_fail(self) -> None:
        """
        Fuzz Test to check argument error detection.
        """
        def generate_random_string(min_length: int = 1, max_length: int = 20) -> str:
            """Generate a random string of random length."""
            length = random.randint(min_length, max_length)  # Random length between min_length and max_length
            return ''.join(random.choices(string.ascii_letters + string.digits + string.punctuation, k=length))

        def generate_random_string_list(
                min_items: int = 1, max_items: int = 3, min_length: int = 1, max_length: int = 20) -> List[str]:
            """Generate a list of random strings with random length and random number of items."""
            num_items = random.randint(min_items, max_items)  # Random number of items in the list
            return [generate_random_string(min_length, max_length) for _ in range(num_items)]

        # Fuzzing loop
        for i in range(100):
            args = ['python', self.pyfile] + generate_random_string_list()
            result = subprocess.run(args, capture_output=True)
            print(result)

            safely_quit = result.returncode == 0
            detected_error = 'Exception' in result.stdout.decode('utf-8') or 'Warning' in result.stdout.decode('utf-8')

            if not any(help_option in args for help_option in ('-h', '--help')):
                self.assertTrue(safely_quit, f'Failed with args: {args}. Full file path: {abspath(self.pyfile)}')
                self.assertTrue(detected_error, f'Failed with args: {args}. Full file path: {abspath(self.pyfile)}')

    def test_warning(self) -> None:
        """
        Test to check if warning raised about too many argument.
        """
        args = [
            'python',
            self.pyfile, '-t',
            f'{self.dc_dir}Default_Configs/default_auto.json',
            f'{self.dc_dir}Default_Configs/default_manual.json',
            'one too many arguments'
        ]
        result = subprocess.run(args, capture_output=True)
        print(result)

        safely_quit = result.returncode == 0
        detected_too_many_args = 'Warning' in result.stdout.decode('utf-8')

        self.assertTrue(safely_quit, f'Failed with args: {args}. Full file path: {abspath(self.pyfile)}')
        self.assertTrue(
            detected_too_many_args, f'Failed with args: {args}. Full file path: {abspath(self.pyfile)}'
        )



"""
Simulation specific utilities that make more sense to keep outside classes.
"""

import random
from typing import Union
from cryptography.hazmat.primitives.hashes import *
from cryptography.hazmat.primitives.asymmetric import rsa, ec


def hash_info(information: str, holder_hash: str) -> str:
    """
    Uses the established hash string to hash whatever information is passed.
    Raises an error if the choice is unsupported.

    Args:
        information: str - Any information as long as it is in string format.
        holder_hash: str - The (assumingly acceptable) hash of choice.

    Returns:
        str: The hashed information if choice is valid.
    """

    if holder_hash == 'sha224':
        digest: Hash = Hash(SHA224())
    elif holder_hash == 'sha256':
        digest: Hash = Hash(SHA256())
    elif holder_hash == 'sha384':
        digest: Hash = Hash(SHA384())
    elif holder_hash == 'sha512':
        digest: Hash = Hash(SHA512())
    elif holder_hash == 'sha3_224':
        digest: Hash = Hash(SHA3_224())
    elif holder_hash == 'sha3_256':
        digest: Hash = Hash(SHA3_256())
    elif holder_hash == 'sha3_384':
        digest: Hash = Hash(SHA3_384())
    elif holder_hash == 'sha3_512':
        digest: Hash = Hash(SHA3_512())
    elif holder_hash == 'blake2b':
        digest: Hash = Hash(BLAKE2b(64))
    elif holder_hash == 'blake2s':
        digest: Hash = Hash(BLAKE2s(32))
    else:
        raise ValueError(f"Unsupported hash: {holder_hash}")

    digest.update(information.encode('utf-8'))
    return digest.finalize().hex()


def get_random_country() -> str:
    """
    Retrieves a random country from a list of countries.

    Returns:
        str: The random country chosen.
    """

    countries: tuple = (
        "Afghanistan", "Albania", "Algeria", "Andorra", "Angola", "Antigua and Barbuda", "Argentina", "Armenia",
        "Australia", "Austria", "Azerbaijan", "The Bahamas", "Bahrain", "Bangladesh", "Barbados", "Belarus", "Belgium",
        "Belize", "Benin", "Bhutan", "Bolivia", "Bosnia and Herzegovina", "Botswana", "Brazil", "Brunei", "Bulgaria",
        "Burkina Faso", "Burundi", "Cabo Verde", "Cambodia", "Cameroon", "Canada", "Central African Republic", "Chad",
        "Chile", "China", "Colombia", "Comoros", "Congo, Democratic Republic of the", "Congo, Republic of the",
        "Costa Rica", "Côte d’Ivoire", "Croatia", "Cuba", "Cyprus", "Czech Republic", "Denmark", "Djibouti",
        "Dominica", "Dominican Republic", "East Timor (Timor-Leste)", "Ecuador", "Egypt", "El Salvador",
        "Equatorial Guinea", "Eritrea", "Estonia", "Eswatini", "Ethiopia", "Fiji", "Finland", "France", "Gabon",
        "The Gambia", "Georgia", "Germany", "Ghana", "Greece", "Grenada", "Guatemala", "Guinea", "Guinea-Bissau",
        "Guyana", "Haiti", "Honduras", "Hungary", "Iceland", "India", "Indonesia", "Iran", "Iraq", "Ireland", "Israel",
        "Italy", "Jamaica", "Japan", "Jordan", "Kazakhstan", "Kenya", "Kiribati", "Korea, North", "Korea, South",
        "Kosovo", "Kuwait", "Kyrgyzstan", "Laos", "Latvia", "Lebanon", "Lesotho", "Liberia", "Libya", "Liechtenstein",
        "Lithuania", "Luxembourg", "Madagascar", "Malawi", "Malaysia", "Maldives", "Mali", "Malta", "Marshall Islands",
        "Mauritania", "Mauritius", "Mexico", "Micronesia, Federated States of", "Moldova", "Monaco", "Mongolia",
        "Montenegro", "Morocco", "Mozambique", "Myanmar (Burma)", "Namibia", "Nauru", "Nepal", "Netherlands",
        "New Zealand", "Nicaragua", "Niger", "Nigeria", "North Macedonia", "Norway", "Oman", "Pakistan", "Palau",
        "Panama", "Papua New Guinea", "Paraguay", "Peru", "Philippines", "Poland", "Portugal", "Qatar", "Romania",
        "Russia", "Rwanda", "Saint Kitts and Nevis", "Saint Lucia", "Saint Vincent and the Grenadines", "Samoa",
        "San Marino", "Sao Tome and Principe", "Saudi Arabia", "Senegal", "Serbia", "Seychelles", "Sierra Leone",
        "Singapore", "Slovakia", "Slovenia", "Solomon Islands", "Somalia", "South Africa", "Spain", "Sri Lanka",
        "Sudan", "South Sudan", "Suriname", "Sweden", "Switzerland", "Syria", "Taiwan", "Tajikistan", "Tanzania",
        "Thailand", "Togo", "Tonga", "Trinidad and Tobago", "Tunisia", "Turkey", "Turkmenistan", "Tuvalu", "Uganda",
        "Ukraine", "United Arab Emirates", "United Kingdom", "United States", "Uruguay", "Uzbekistan", "Vanuatu",
        "Vatican City", "Venezuela", "Vietnam", "Yemen", "Zambia", "Zimbabwe"
    )

    return random.choice(countries)


def get_random_division() -> str:
    """
    Retrieves a random business division from a list of divisions.

    Returns:
        str: The random business division chosen.
    """

    business_units: tuple = (
        "Human Resources", "Finance", "Accounting", "Marketing", "Sales", "Operations", "IT (Information Technology)",
        "Research and Development (R&D)", "Customer Service", "Legal", "Procurement", "Supply Chain Management",
        "Production", "Quality Assurance", "Public Relations (PR)", "Business Development", "Strategy and Planning",
        "Compliance", "Data Analytics", "Training and Development", "Facilities Management", "Risk Management",
        "Corporate Social Responsibility (CSR)", "Internal Audit", "Engineering", "Product Management",
        "Executive Management", "Investor Relations", "Logistics", "Security", "Operations Management"
    )

    return random.choice(business_units)


def get_curve(holder_curve: str) -> ec.EllipticCurve:
    """
    Takes a passed curve and returns a curve object, or raises an error if the curve is unsupported.

    Args:
        holder_curve: str - The passed curve choice.

    Returns:
        ec.EllipticCurve: The elliptic curve object.
    """

    if holder_curve == 'secp256r1':
        return ec.SECP256R1()
    elif holder_curve == 'secp384r1':
        return ec.SECP384R1()
    elif holder_curve == 'secp521r1':
        return ec.SECP521R1()
    elif holder_curve == 'secp224r1':
        return ec.SECP224R1()
    elif holder_curve == 'secp192r1':
        return ec.SECP192R1()
    elif holder_curve == 'secp256k1':
        return ec.SECP256K1()
    else:
        raise ValueError(f"Unsupported curve: {holder_curve}")


def create_private_key(holder_encrypt_alg: dict) -> Union[rsa.RSAPrivateKey, ec.EllipticCurvePrivateKey]:
    """
    Takes encryption algorithm parameters and returns either a private key or Nothing.

    Args:
        holder_encrypt_alg: dict - dictionary of encryption algorithm settings

    Returns:
        Union[rsa.RSAPrivateKey, ec.EllipticCurvePrivateKey, None]: The private key or lack of one.
    """

    params: dict = holder_encrypt_alg['params']
    try:
        if holder_encrypt_alg['alg'] == 'rsa':
            return rsa.generate_private_key(public_exponent=params['pub_exp'], key_size=params['key_size'])
        elif holder_encrypt_alg['alg'] == 'ecc':
            curve: ec.EllipticCurve = get_curve(params['curve'])
            return ec.generate_private_key(curve=curve)
    except ValueError as e:
        assert False, (
            'Incorrect value detected during Holder creation.\n' 
            f'\t   Message: {e}.\n'
            '\t   Please check automatic or manual configuration files that contain the passed value.'
        )



"""
Module used for defining the network class and it's functionality.
"""

# Relative pathing from project root
import sys
import datetime
from os import makedirs
from os.path import abspath, dirname, join, exists
from typing import Union, List, Dict

script_dir = dirname(abspath(__file__))

if script_dir in ['PKI_Practice', 'PKI Practice', 'app']:
    sys.path.append(abspath(script_dir))
elif script_dir == 'PKIPractice':
    sys.path.append(abspath(join(script_dir, '..')))
else:
    sys.path.append(abspath(join(script_dir, '../..')))

from PKIPractice.Simulation.Holder import PKIHolder


class PKIHub:
    """
    The "infrastructure" of the network. Will be used to direct communications between holders.

    Attributes:
        outer_network (PKINetwork): The network object to access information and logging.
        mirror_store (dict): The "mirror" of the environment hierarchy in literal name.
        loc_store (dict): The location of every holder in the environment.
        send_store (dict): Where to send information for each holder, split into sending "up" or "down" the PKI
            hierarchy.

    Methods:
        receive_log(holder_cat, holder_success, holder_act, holder_output, holder_name, holder_message) -> None:
            Takes a log from a holder and sends it to network logs.
    """
    def __init__(self, network, network_store: Dict[int, List[PKIHolder]]) -> None:
        self.outer_network = network

        # Create a mirror of network and location store
        self.mirror_store: dict = {}
        self.loc_store: dict = {}  # Sender = key, Receiver = value

        for level, holders in network_store.items():
            self.mirror_store[level] = []
            for holder in holders:
                # Add the address of the holder to the store
                self.mirror_store[level].append(holder.get_addr())

                # Add the location of the holder
                self.loc_store[holder.get_addr()] = (level, len(self.mirror_store[level]) - 1)

                # Set the hub for the holder
                holder.set_hub_conn(self)

        # Use mirror store to decide who to send things to.
        self.send_store: dict = {'up': {}, 'down': {}}
        for level in self.mirror_store.keys():
            # Skip first level
            if level == 1:
                continue

            prev_level = self.mirror_store[level-1]
            prev_level_index = 0
            PREV_LEVEL_MAX = len(prev_level)

            for holder_addr in self.mirror_store[level]:
                # Record where the holder will send the message to
                self.send_store['up'][holder_addr] = prev_level[prev_level_index]

                # Record where the receiver can send the message back
                if prev_level[prev_level_index] in self.send_store['down'].keys():
                    self.send_store['down'][prev_level[prev_level_index]].append(holder_addr)
                else:
                    self.send_store['down'][prev_level[prev_level_index]] = [holder_addr]

                prev_level_index += 1
                if prev_level_index == PREV_LEVEL_MAX:
                    prev_level_index = 0

    def receive_log(self, holder_cat: str, holder_success: bool, holder_act: str, holder_output: str,
                    holder_name: str, holder_message: str) -> None:
        """
        Receives a log from a holder and saved it to network logs.
        """
        self.outer_network.log_event(
            holder_cat, holder_success, 'Holder', holder_act, holder_output, holder_name, holder_message
        )

    def forward_message(self) -> None:
        """Sends a message from one holder to another."""
        ...


class PKINetwork:
    """
    Represents a Public Key Infrastructure (PKI) network, including its configuration,
    hierarchy, and associated holders. This class provides functionality for managing
    the network, logging events, and filling configuration gaps.

    Attributes:
        network_name (str): Unique identifier for the network.
        log_save_fp (str): Where the logs will be saved to on storage drive.
        network_level_count (int): Total levels in the network hierarchy.
        network_count_by_level (List[int]): Number of holders at each level in the hierarchy.
        network_total_count (int): Total number of holders in the network.
        env_uid_hash (str): Unique identifier hash for the network environment.
        env_sig_hash (str): Signature hash used in the network.
        env_encrypt_alg (dict): Encryption algorithms supported by the network.
        env_revoc_probs (List[float]): Revocation probabilities by network level.
        env_cert_valid_durs (List[str]): Certificate validity durations by level.
        env_cache_durs (List[str]): Cache durations by network level.
        env_cooldown_durs (List[str]): Cooldown durations by network level.
        env_timeout_durs (List[str]): Timeout durations by network level.
        network (dict): Dictionary mapping network levels to their associated holders.
        network_log (List[str]): Log of events that occurred in the network.
        network_hub (PKIHub): Central hub for managing network communications.

    Methods:
        log_event(message) -> None:
            Prints the event that had happened and saves the information to network logs.
        save_logs() -> None:
            Saves the logs generated by the network to a CSV file.
        add_to_network(holder_name, holder_config, auto_config) -> bool:
            Adds a new holder to the network and returns a success status.
        get_network() -> List[PKIHolder]:
            Returns a flat list of all holders in network ordered by highest level.
        set_root_certificates() -> None:
            Sets up the certificate of root holders.
    """
    def __init__(self, name: str, auto_config: Union[dict, None], manual_config: Union[dict, None] = None) -> None:
        # Unique identifier
        self.network_name: str = name

        # Log filepath
        self.log_save_fp: str = auto_config['log_save_filepath']

        # Network counts
        self.network_level_count: int = auto_config['level_count']
        self.network_count_by_level: List[int] = auto_config['count_by_level']
        self.network_total_count: int = 0

        # Environment variables
        self.env_uid_hash: str = auto_config['uid_hash']
        self.env_sig_hash: str = auto_config['sig_hash']
        self.env_encrypt_alg: dict = auto_config['encrypt_alg']
        self.env_revoc_probs: List[float] = auto_config['revoc_probs']
        self.env_cert_valid_durs: List[str] = auto_config['cert_valid_durs']
        self.env_cache_durs: List[str] = auto_config['cache_durs']
        self.env_cooldown_durs: List[str] = auto_config['cooldown_durs']
        self.env_timeout_durs: List[str] = auto_config['timeout_durs']

        # Network hierarchy
        self.network: dict = {}
        for i in range(self.network_level_count):
            self.network[i+1] = []

        self.network_log: List[str] = ['ID, Timestamp, Category, Success, Subject, Act, Output, Origin, Message\n']

        # Log events that have already happened
        self.log_event(
            'Operations', True, 'Network', 'Initialization', 'Network',
            self.network_name, f'Network {self.network_name} created.'
        )
        self.log_event(
            'Operations', True, 'Network', 'Initialization', 'Variables',
            self.network_name, 'Environmental variables set.'
        )
        self.log_event(
            'Operations', True, 'Network', 'Initialization', 'Hierarchy',
            self.network_name, 'Empty network hierarchy created.'
        )
        self.log_event(
            'Operations', True, 'Network', 'Initialization', 'Log',
            self.network_name, 'Network log created and started.'
        )

        # Manual configuration
        if manual_config is not None:
            for holder_name, holder_config in manual_config.items():
                result: bool = self.add_to_network(holder_name, holder_config, auto_config)
                if result:
                    self.log_event(
                        'Operations', True, 'Network', 'Addition', holder_name,
                        self.network_name, f'Holder {holder_name} added to network.'
                    )
                    self.network_total_count += 1
                else:
                    self.log_event(
                        'Operations', False, 'Network', 'Omission', holder_name,
                        self.network_name, f'Invalid location configuration. {holder_name} was ignored.'
                    )

        # Filling in gaps
        auto_holder_count = 1
        for i in range(self.network_level_count):
            while len(self.network[i+1]) < self.network_count_by_level[i]:
                self.add_to_network(
                    f'holder_l{i+1}_c{auto_holder_count}',
                    {'location': {'level': i+1}},
                    auto_config
                )
                self.log_event(
                    'Operations', True, 'Network', 'Completion',
                    f'holder_l{i+1}_c{auto_holder_count}', self.network_name,
                    f'Gap found. Filler Holder #{auto_holder_count} at level {i+1} added to network.'
                )
                auto_holder_count += 1
                self.network_total_count += 1

        # Network hub
        self.network_hub = PKIHub(self, self.network)
        self.log_event(
            'Operations', True, 'Network', 'Initialization', 'Hub', self.network_name,
            'Network hub and connection information created.'
        )

    def log_event(self, category: str, is_success: bool, subject: str, act: str, output: str, origin: str,
                  message: str) -> None:
        """
        Takes a message, prints it, and saves its entry to network log with other elements.

        Categories that are used are:
            Operations - General Network activities
            PKI - Activities specific to PKI Practices.
            Communication - Activities specific to Communication.

        Args:
            category: str - The category of the message. One word.
            is_success: bool - The success status of the message.
            subject: str - The subject of the message. One word.
            act: str - The noun as an act the subject did. One word.
            output: str - The thing that resulted from the act. One word.
            origin: str - Who called the logging function. One word.
            message: str - String value to save.
        """

        id_num = len(self.network_log) - 1
        timestamp = datetime.datetime.now()
        entry = f'{id_num}, {timestamp}, {category}, {is_success}, {subject}, {act}, {output}, {origin}, {message}\n'

        print(message)
        self.network_log.append(entry)

    def save_logs(self) -> None:
        """
        Saves the logs to a csv file.
        """

        # Create one last log event
        self.log_event(
            'Operations', True, 'Network', 'Retention', 'CSV', self.network_name,
            f'Logs of network saved to {self.log_save_fp}.'
        )

        # Make a directory if it doesn't exist
        if not exists(dirname(self.log_save_fp)):
            makedirs(dirname(self.log_save_fp))

        # Save the log file
        with open(self.log_save_fp, 'w') as log_file:
            log_file.writelines(self.network_log)

    def add_to_network(self, holder_name: str, holder_config: dict, auto_config: dict) -> bool:
        """
        Takes a given holder name, holder configuration dictionary, and auto_config elements and creates a holder.
        Adds the holder to the growing network.

        Args:
            holder_name: str - The name of the manually created holder.
            holder_config: dict - The configuration settings of the holder.
            auto_config: dict - The configuration settings of the environment.

        Returns:
            bool - Success status on operation.
        """

        # Check if location is valid
        proper_keys: bool = all(
            isinstance(holder_config['location'][key], int) for key in holder_config['location'].keys()
        )
        enough_keys: bool = len(holder_config['location'].keys()) == 1

        if not proper_keys or not enough_keys:
            print(f'Invalid location configuration. {holder_name} configuration will be ignored.')
            return False

        # Create holder
        holder: PKIHolder = PKIHolder(holder_name, holder_config, auto_config)

        # Add holder to network
        level = holder_config['location']['level']
        self.network[level].append(holder)

        return True

    def get_network(self) -> List[PKIHolder]:
        """
        Returns a flat list of all holders in the environment in order of level.

        Returns:
            List[PKIHolder]: A list of all holder objects.
        """

        flat_network = [holder for level in self.network.values() for holder in level]
        return flat_network

    def set_root_certificates(self) -> None:
        """
        Sets up the root certificates for the top level them passes them all to all holders.
        """

        # Generate self-signed certificates
        for root_holder in self.network[1]:
            root_holder.gen_self_cert()



"""
Top level entry point for the PyPKIPractice Program.

Methods:
    run_pki_practice - command line entry point to start the program and take in arguments.
"""

from .Utilities.CLIUtils import start_program


def run_pki_practice() -> None:
    """
    Command line entry point to start the program and take in arguments.
    """

    start_program()



"""
This file contains functions for parsing configuration files and validating their input.
"""

import sys
from typing import Union

# Specifying tomli library version
if sys.version_info[1] >= 11:
    import tomllib
else:
    import tomli

# Specifying yaml library version
if sys.version_info[1] > 9:
    import yaml

from xml.etree import ElementTree
import json
import re
from .EnumUtils import *


def parse_child_tags(xml_string: str) -> dict:
    """
    Parses the child tags of an XML string.

    Args:
        xml_string (str): The XML string to be parsed.

    Returns:
        dict: A dictionary containing the parsed child tags.
    """

    root = ElementTree.fromstring(xml_string)
    result = {}
    for child in root:
        if len(child) == 0 and child.tag not in result:  # New tag not come across inside parent tag
            result[child.tag] = child.text
        elif len(child) == 0 and isinstance(result[child.tag], str):  # Has come across a tag for the second time
            result[child.tag] = [result[child.tag], child.text]
        elif len(child) == 0 and isinstance(result[child.tag], list):  # Has come across a tag more than twice
            result[child.tag].append(child.text)
        else:  # Found a new parent tag
            result[child.tag] = parse_child_tags(ElementTree.tostring(child))
    return result


def adjust_types_auto(settings: dict) -> Union[dict, None]:
    """
    Adjusts the data types and format of the settings dictionary.

    Args:
        settings (dict): The dictionary to be modified and wrangled.

    Returns:
        dict: The tailored dictionary.
    """

    try:
        settings['level_count'] = int(settings['level_count'])
        settings['count_by_level'] = list(map(int, settings['count_by_level']))
        settings['uid_hash'] = settings['uid_hash'].lower().replace('-', '_')
        settings['sig_hash'] = settings['sig_hash'].lower().replace('-', '_')
        settings['encrypt_alg']['alg'] = settings['encrypt_alg']['alg'].lower()

        if settings['encrypt_alg']['alg'] == 'rsa':
            settings['encrypt_alg']['params']['pub_exp'] = int(settings['encrypt_alg']['params']['pub_exp'])
            settings['encrypt_alg']['params']['key_size'] = int(settings['encrypt_alg']['params']['key_size'])
        if settings['encrypt_alg']['alg'] == 'ecc':
            settings['encrypt_alg']['params']['curve'] = settings['encrypt_alg']['params']['curve'].lower()

        settings['revoc_probs'] = list(map(float, settings['revoc_probs']))
    except (KeyError, TypeError, ValueError) as e:
        topic = re.search(r"'(.*?)'", str(e)).group(1)
        print(f'Either "{topic}" is a missing key word, or "{topic}" is not the correct data type.\nIf the former is '
              'true, please add it to the configuration file.\nIf the latter is true, look for where it is used in '
              'the autoconfiguration file, as that is likely the problem.')
        return None

    return settings


def validate_settings_auto(settings: dict) -> bool:
    """
    Validates the settings dictionary.

    Args:
        settings (dict): The dictionary to be validated.

    Returns:
        bool: True if the settings are valid, False otherwise.
    """

    # Checking the existence and lengths of lists
    try:
        for setting in [
            settings['count_by_level'], settings['revoc_probs'], settings['cert_valid_durs'],
            settings['cache_durs'], settings['cooldown_durs'], settings['timeout_durs']
        ]:
            if not isinstance(setting, list):
                print(f'The value {setting} is not a list. Please fix this in the autoconfiguration file.')
                return False
            if len(setting) != settings['level_count']:
                print(f'The number of values in the list {setting} must match the level_count parameter. '
                      'Please fix this in the autoconfiguration file.')
                return False
    except KeyError as e:
        print(f'{e} is a missing key required in the autoconfiguration file. Please add it.')
        return False

    # Checking existence of untouched strings
    if not settings['uid_hash'] or not settings['sig_hash'] or not settings['encrypt_alg']['alg']:
        print('uid_hash, sig_hash, and encrypt_alg.alg are missing from the autoconfiguration file. Please add them.')
        return False

    # Checking durations for correct formats
    for dur in settings['cert_valid_durs']:
        if not (re.match(r'^[0-9]+:[0-9]{2}:[0-9]{2}$', dur) or dur == 'none'):
            print(f'"{dur}" is not a valid input for cert_valid_durs. Please fix this in the autoconfiguration file.')
            return False

    for dur in settings['cache_durs']:
        if not (re.match(r'^[0-9]{2}:[0-9]{2}$', dur) or dur == 'none'):
            print(f'"{dur}" is not a valid input for cache_durs. Please fix this in the autoconfiguration file.')
            return False

    for dur in settings['cooldown_durs']:
        if not (re.match(r'^[0-9]+$', dur) or dur == 'none'):
            print(f'"{dur}" is not a valid input for cooldown_durs. Please fix this in the autoconfiguration file.')
            return False

    for dur in settings['timeout_durs']:
        if not (re.match(r'^[0-9]+$', dur) or dur == 'none'):
            print(f'"{dur}" is not a valid input for timeout_durs. Please fix this in the autoconfiguration file.')
            return False

    # Checking if revoc_probs are between 0 and 1 inclusive
    for prob in settings['revoc_probs']:
        if prob < 0.0 or prob > 1.0:
            print(f'"{prob}" is not a valid input for revoc_probs and must be between 0 and 1. '
                  'Please fix this in the autoconfiguration file.')
            return False

    # Checking if parameters for hashing and encryption are valid
    if not has_value(SUPPORTED_HASH_ALGS, settings['uid_hash']):
        print(f'"{settings["uid_hash"]}" is not a valid input for uid_hash. Please fix this in the autoconfiguration ' 
              'file.')
        return False

    if not has_value(SUPPORTED_HASH_ALGS, settings['sig_hash']):
        print(f'"{settings["sig_hash"]}" is not a valid input for sig_hash. Please fix this in the autoconfiguration '
              'file.')
        return False

    if not has_value(SUPPORTED_ENCRYPT_ALGS, settings['encrypt_alg']['alg']):
        print(f'"{settings["encrypt_alg"]["alg"]}" is not a valid input for encrypt_alg.alg. Please fix this in the '
              'autoconfiguration file.')
        return False

    if settings['encrypt_alg']['alg'] == 'ecc':
        if not has_value(SUPPORTED_ECC_CURVES, settings['encrypt_alg']['params']['curve']):
            print(f'"{settings["encrypt_alg"]["params"]["curve"]}" is not a valid input for '
                  'encrypt_alg.params.curve. Please fix this in the autoconfiguration file.')
            return False

    # Checking if a proper filepath is passed for the program.
    if 'log_save_filepath' not in settings.keys():
        print('The filepath for saving log files is not existent. Please add that to the autoconfiguration file.')
        return False

    if len(settings['log_save_filepath']) < 7:
        print('The filepath for saving log files is too short to be valid. '
              'File paths need to be at least 7 characters long. '
              'Please fix that in the autoconfiguration file.')
        return False

    if settings['log_save_filepath'][-4:] != '.csv':
        print('Invalid log file path found. '
              'Please ensure that the filepath you pass to the network has the ".csv" extension at the end. '
              'Logs are saved as CSV files for convenience. '
              'Please fix this in the autoconfiguration file.')
        return False

    return True


def parse_config_auto(filepath: str) -> Union[dict, None]:
    """
    Parses an autoconfiguration file given its file path.

    Args:
        filepath (str): The path to the configuration file to be parsed.

    Returns:
        dict: A dictionary containing the parsed configuration data.
    """

    settings: Union[dict, None] = None

    # Check file type
    assert any(ext in filepath for ext in ['.yaml', '.yml', '.json', '.xml', '.toml']), (
        'Invalid autoconfiguration configuration file provided.\n'
        '\t   Please provide a configuration file that is a YAML, JSON, XML, or TOML file.\n'
        '\t   Look in the Default_Configs folder for examples.\n'
    )

    # File type tree
    try:
        if filepath.endswith('.yaml') or filepath.endswith('.yml'):
            assert sys.version_info[1] > 9, (
                'Using YAML files for configuration requires Python 3.10 or higher.\n'
                '\t   Please update your Python version or use a different file format for configuration.\n'
                '\t   Other supported examples are JSON, XML, or TOML.\n'
                '\t   Look in the Default_Configs folder for examples.\n'
            )
            with open(filepath, 'r') as file:
                settings = yaml.load(file, Loader=yaml.Loader)
        elif filepath.endswith('.json'):
            with open(filepath, 'r') as file:
                settings = json.load(file)
        elif filepath.endswith('.xml'):
            with open(filepath, 'r') as file:
                settings = parse_child_tags(file.read())
        elif filepath.endswith('.toml'):
            if sys.version_info[1] >= 11:
                with open(filepath, 'rb') as file:
                    settings = tomllib.load(file)
            else:
                with open(filepath, 'rb') as file:
                    settings = tomli.load(file)
    except Exception as e:
        print(f'Ingestion libraries experienced an error: "{str(e).title()}"')
        return settings

    # Type adjustment
    settings = adjust_types_auto(settings)
    assert settings is not None, (
        'Ingested autoconfiguration settings were not able to be adjusted due to incorrect configuration format.\n'
        '\t   Please ensure your configuration file is correctly created.\n'
        '\t   Use the default configuration file as a template.\n'
    )

    # Settings validation
    assert validate_settings_auto(settings) is True, (
        'Ingested autoconfiguration settings were not found to be valid.\n'
        '\t   Please ensure your configuration file is correctly created.\n'
        '\t   Use the default configuration file as a template.\n'
    )

    return settings


def search_for_typecast_manual(settings: dict) -> Union[dict, None]:
    """
    Adjusts the data types and format of the settings dictionary.

    Args:
        settings (dict): The dictionary to be modified and wrangled.

    Returns:
        dict: The tailored dictionary.
    """

    try:
        for key, value in settings.items():
            if key in ['level', 'pub_exp', 'key_size']:
                settings[key] = int(value)
            elif key in ['revoc_prob']:
                settings[key] = float(value)
            elif isinstance(value, dict):
                settings[key] = search_for_typecast_manual(value)
                assert settings[key] is not None, (
                    'Ingested manual configuration settings were not able to be adjusted '
                    'due to incorrect configuration format.\n'
                    '\t   Please ensure your configuration file is correctly created.\n'
                    '\t   Use the default configuration file as a template.\n'
                )
        return settings
    except (KeyError, TypeError, ValueError) as e:
        print(e)
        match = re.search(r"'(.*?)'", str(e))
        if match:
            problematic_key = match.group(1)
            print(f'Look for where "{problematic_key}" is used in the manual configuration file, '
                  'as that is likely the problem.')


def parse_config_manual(filepath: str) -> Union[dict, None]:
    """
    Parses a manual configuration file given its file path.

    Args:
        filepath (str): The path to the configuration file to be parsed.

    Returns:
        dict: A dictionary containing the parsed configuration data.
    """

    settings: Union[dict, None] = None

    # Check file type
    assert any(ext in filepath for ext in ['.yaml', '.yml', '.json', '.xml', '.toml']), (
        'Invalid manual configuration file provided.\n'
        '\t   Please provide a configuration file that is a YAML, JSON, XML, or TOML file.\n'
        '\t   Look in the Default_Configs folder for examples.\n'
    )

    # File type tree
    try:
        if filepath.endswith('.yaml') or filepath.endswith('.yml'):
            assert sys.version_info[1] > 9, (
                'Using YAML files for configuration requires Python 3.10 or higher.\n'
                '\t   Please update your Python version or use a different file format for configuration.\n'
                '\t   Other supported examples are JSON, XML, or TOML.\n'
                '\t   Look in the Default_Configs folder for examples.\n'
            )
            with open(filepath, 'r') as file:
                settings = yaml.load(file, Loader=yaml.Loader)
        elif filepath.endswith('.json'):
            with open(filepath, 'r') as file:
                settings = json.load(file)
        elif filepath.endswith('.xml'):
            with open(filepath, 'r') as file:
                settings = parse_child_tags(file.read())
        elif filepath.endswith('.toml'):
            if sys.version_info[1] >= 11:
                with open(filepath, 'rb') as file:
                    settings = tomllib.load(file)
            else:
                with open(filepath, 'rb') as file:
                    settings = tomli.load(file)
    except Exception as e:
        print(f'Ingestion libraries experienced an error: "{str(e).title()}"')
        return settings

    # Type adjustment
    settings = search_for_typecast_manual(settings)
    assert settings is not None, (
        'Ingested manual configuration settings were not able to be adjusted due to unparsable configuration params.\n'
        '\t   Please ensure your configuration file is correctly created.\n'
        '\t   Use the default configuration file as a template.\n'
    )

    return settings
